#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest

from dasbus.specification import DBusSpecificationParser, DBusSpecification, \
    DBusSpecificationError


class SpecificationTestCase(unittest.TestCase):

    def test_from_members(self):
        """Test a specification created from members."""
        method = DBusSpecification.Method(
            name="Method",
            interface_name="A",
            in_type="()",
            out_type="(s)"
        )

        signal = DBusSpecification.Signal(
            name="Signal",
            interface_name="B",
            type="(ii)"
        )

        prop = DBusSpecification.Property(
            name="Property",
            interface_name="B",
            readable=True,
            writable=False,
            type="i"
        )

        specification = DBusSpecification()
        specification.add_member(method)
        specification.add_member(signal)
        specification.add_member(prop)

        with self.assertRaises(DBusSpecificationError) as cm:
            specification.get_member("A", "Invalid")

        self.assertEqual(
            "DBus specification has no member 'A.Invalid'.",
            str(cm.exception)
        )

        with self.assertRaises(DBusSpecificationError) as cm:
            specification.get_member("Invalid", "Method")

        self.assertEqual(
            "DBus specification has no member 'Invalid.Method'.",
            str(cm.exception)
        )

        self.assertEqual(specification.interfaces, ["A", "B"])
        self.assertEqual(specification.members, [method, signal, prop])
        self.assertEqual(specification.get_member("A", "Method"), method)
        self.assertEqual(specification.get_member("B", "Signal"), signal)
        self.assertEqual(specification.get_member("B", "Property"), prop)

    def test_from_xml(self):
        """Test a specification created from XML."""
        xml = '''
        <node>
           <interface name="ComplexA">
                <method name="MethodA">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <property access="read" name="PropertyA" type="d"/>
                <signal name="SignalA">
                   <arg direction="out" name="x" type="i"/>
                </signal>
            </interface>
            <interface name="ComplexB">
                <method name="MethodB">
                    <arg direction="in" name="a" type="s"/>
                    <arg direction="in" name="b" type="ad"/>
                    <arg direction="in" name="c" type="i"/>
                    <arg direction="out" name="return" type="i"/>
                </method>
                <property access="read" name="PropertyB" type="d"/>
                <signal name="SignalB">
                    <arg direction="out" name="x" type="b"/>
                    <arg direction="out" name="y" type="d"/>
                    <arg direction="out" name="z" type="(ii)"/>
                </signal>
            </interface>
        </node>
        '''
        specification = DBusSpecification.from_xml(xml)
        self.assertEqual(specification.interfaces, [
            'org.freedesktop.DBus.Introspectable',
            'org.freedesktop.DBus.Peer',
            'org.freedesktop.DBus.Properties',
            'ComplexA',
            'ComplexB',
        ])

        self.assertEqual(specification.members, [
            DBusSpecification.Method(
                name='Introspect',
                interface_name='org.freedesktop.DBus.Introspectable',
                in_type=None,
                out_type='(s)'
            ),
            DBusSpecification.Method(
                name='Ping',
                interface_name='org.freedesktop.DBus.Peer',
                in_type=None,
                out_type=None
            ),
            DBusSpecification.Method(
                name='GetMachineId',
                interface_name='org.freedesktop.DBus.Peer',
                in_type=None,
                out_type='(s)'
            ),
            DBusSpecification.Method(
                name='Get',
                interface_name='org.freedesktop.DBus.Properties',
                in_type='(ss)',
                out_type='(v)'
            ),
            DBusSpecification.Method(
                name='GetAll',
                interface_name='org.freedesktop.DBus.Properties',
                in_type='(s)',
                out_type='(a{sv})'
            ),
            DBusSpecification.Method(
                name='Set',
                interface_name='org.freedesktop.DBus.Properties',
                in_type='(ssv)',
                out_type=None
            ),
            DBusSpecification.Signal(
                name='PropertiesChanged',
                interface_name='org.freedesktop.DBus.Properties',
                type='(sa{sv}as)'
            ),
            DBusSpecification.Method(
                name='MethodA',
                interface_name='ComplexA',
                in_type=None,
                out_type='(i)'
            ),
            DBusSpecification.Property(
                name='PropertyA',
                interface_name='ComplexA',
                readable=True,
                writable=False,
                type='d'
            ),
            DBusSpecification.Signal(
                name='SignalA',
                interface_name='ComplexA',
                type='(i)'
            ),
            DBusSpecification.Method(
                name='MethodB',
                interface_name='ComplexB',
                in_type='(sadi)',
                out_type='(i)'
            ),
            DBusSpecification.Property(
                name='PropertyB',
                interface_name='ComplexB',
                readable=True,
                writable=False,
                type='d'
            ),
            DBusSpecification.Signal(
                name='SignalB',
                interface_name='ComplexB',
                type='(bd(ii))'
            )
        ])


class SpecificationParserTestCase(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None

    def _compare(self, xml, expected_members):
        """Compare members of the specification."""
        specification = DBusSpecification()
        DBusSpecificationParser._parse_xml(specification, xml)
        self.assertEqual(specification.members, expected_members)

    def test_method(self):
        """Test XML with methods."""
        xml = '''
        <node>
            <interface name="Interface">
                <method name="Method1"/>
                <method name="Method2">
                    <arg direction="in" name="x" type="i"/>
                </method>
                <method name="Method3">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <method name="Method4">
                    <arg direction="in" name="x" type="ad"/>
                    <arg direction="in" name="y" type="h"/>
                    <arg direction="out" name="return" type="(ib)"/>
                </method>
            </interface>
        </node>
        '''
        self._compare(xml, [
            DBusSpecification.Method(
                name='Method1',
                interface_name='Interface',
                in_type=None,
                out_type=None
            ),
            DBusSpecification.Method(
                name='Method2',
                interface_name='Interface',
                in_type='(i)',
                out_type=None
            ),
            DBusSpecification.Method(
                name='Method3',
                interface_name='Interface',
                in_type=None,
                out_type='(i)'
            ),
            DBusSpecification.Method(
                name='Method4',
                interface_name='Interface',
                in_type='(adh)',
                out_type='((ib))'
            )
        ])

    def test_property(self):
        """Test XML with a property."""
        xml = '''
        <node>
            <interface name="Interface">
                <property name="Property" type="i" access="readwrite" />
            </interface>
        </node>
        '''
        self._compare(xml, [
            DBusSpecification.Property(
                name='Property',
                interface_name='Interface',
                readable=True,
                writable=True,
                type='i'
            )
        ])

    def test_property_readonly(self):
        """Test readonly property."""
        xml = '''
        <node>
            <interface name="Interface">
                <property name="Property" type="i" access="read" />
            </interface>
        </node>
        '''
        self._compare(xml, [
            DBusSpecification.Property(
                name='Property',
                interface_name='Interface',
                readable=True,
                writable=False,
                type='i'
            )
        ])

    def test_property_writeonly(self):
        """Test writeonly property."""
        xml = '''
        <node>
            <interface name="Interface">
                <property name="Property" type="i" access="write" />
            </interface>
        </node>
        '''
        self._compare(xml, [
            DBusSpecification.Property(
                name='Property',
                interface_name='Interface',
                readable=False,
                writable=True,
                type='i'
            )
        ])

    def test_simple_signal(self):
        """Test interface with a simple signal."""
        xml = '''
        <node>
            <interface name="Interface">
                <signal name="SimpleSignal"/>
            </interface>
        </node>
        '''
        self._compare(xml, [
            DBusSpecification.Signal(
                name='SimpleSignal',
                interface_name='Interface',
                type=None
            )
        ])

    def test_signal(self):
        """Test interface with signals."""
        xml = '''
        <node>
            <interface name="Interface">
                <signal name="SignalSomething">
                    <arg direction="out" name="x" type="i"/>
                    <arg direction="out" name="y" type="s"/>
                </signal>
                <signal name="SomethingHappened"/>
            </interface>
        </node>
        '''
        self._compare(xml, [
            DBusSpecification.Signal(
                name='SignalSomething',
                interface_name='Interface',
                type='(is)'
            ),
            DBusSpecification.Signal(
                name='SomethingHappened',
                interface_name='Interface',
                type=None
            )
        ])

    def test_ignore(self):
        """Ignore invalid XML elements.."""
        xml = '''
        <node>
           <interface name="InterfaceA">
                <method name="MethodA">
                    <arg direction="out" name="return" type="i"/>
                    <ignored />
                </method>
                <property access="read" name="PropertyA" type="d"/>
                <signal name="SignalA">
                   <ignored />
                </signal>
                <ignored />
            </interface>
            <ignored />
        </node>
        '''
        self._compare(xml, [
            DBusSpecification.Method(
                name='MethodA',
                interface_name='InterfaceA',
                in_type=None,
                out_type='(i)'
            ),
            DBusSpecification.Property(
                name='PropertyA',
                interface_name='InterfaceA',
                readable=True,
                writable=False,
                type='d'
            ),
            DBusSpecification.Signal(
                name='SignalA',
                interface_name='InterfaceA',
                type=None
            )
        ])

        self._compare("<ignored />", [])

    def test_standard_interfaces(self):
        """Test with the standard interfaces."""
        specification = DBusSpecificationParser.parse_specification('<node />')
        self.assertEqual(specification.members, [
            DBusSpecification.Method(
                name='Introspect',
                interface_name='org.freedesktop.DBus.Introspectable',
                in_type=None,
                out_type='(s)'
            ),
            DBusSpecification.Method(
                name='Ping',
                interface_name='org.freedesktop.DBus.Peer',
                in_type=None,
                out_type=None
            ),
            DBusSpecification.Method(
                name='GetMachineId',
                interface_name='org.freedesktop.DBus.Peer',
                in_type=None,
                out_type='(s)'
            ),
            DBusSpecification.Method(
                name='Get',
                interface_name='org.freedesktop.DBus.Properties',
                in_type='(ss)',
                out_type='(v)'
            ),
            DBusSpecification.Method(
                name='GetAll',
                interface_name='org.freedesktop.DBus.Properties',
                in_type='(s)',
                out_type='(a{sv})'
            ),
            DBusSpecification.Method(
                name='Set',
                interface_name='org.freedesktop.DBus.Properties',
                in_type='(ssv)',
                out_type=None
            ),
            DBusSpecification.Signal(
                name='PropertiesChanged',
                interface_name='org.freedesktop.DBus.Properties',
                type='(sa{sv}as)'
            )
        ])



#
# Copyright (C) 2022  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import multiprocessing
import sys
import unittest

from abc import abstractmethod, ABCMeta
from contextlib import contextmanager
from threading import Thread

import gi
gi.require_version("Gio", "2.0")
gi.require_version("GLib", "2.0")
from gi.repository import GLib, Gio


def run_loop(timeout=3):
    """Run an event loop for the specified timeout.

    If any of the events fail or there are some pending
    events after the timeout, raise AssertionError.

    :param int timeout: a number of seconds
    """
    loop = GLib.MainLoop()

    def _kill_loop():
        loop.quit()
        return False

    GLib.timeout_add_seconds(timeout, _kill_loop)

    with catch_errors() as errors:
        loop.run()

    assert not errors, "The loop has failed!"
    assert not loop.get_context().pending()


@contextmanager
def catch_errors():
    """Catch exceptions raised in this context.

    :return: a list of exceptions
    """
    errors = []

    def _handle_error(*exc_info):
        errors.append(exc_info)
        sys.__excepthook__(*exc_info)

    try:
        sys.excepthook = _handle_error
        yield errors
    finally:
        sys.excepthook = sys.__excepthook__


class AbstractDBusTestCase(unittest.TestCase, metaclass=ABCMeta):
    """Test DBus support with a real DBus connection."""

    def setUp(self):
        """Set up the test."""
        self.maxDiff = None

        # Initialize the service and the clients.
        self.service = self._get_service()
        self.clients = []

        # Start a testing bus.
        self.bus = Gio.TestDBus()
        self.bus.up()

        # Create a connection to the testing bus.
        self.bus_address = self.bus.get_bus_address()
        self.message_bus = self._get_message_bus(
            self.bus_address
        )

    @abstractmethod
    def _get_service(self):
        """Get a service."""
        return None

    @classmethod
    @abstractmethod
    def _get_message_bus(cls, bus_address):
        """Get a testing message bus."""
        return None

    @classmethod
    def _get_service_proxy(cls, message_bus, **proxy_args):
        """Get a proxy of the example service."""
        return message_bus.get_proxy(
            "my.testing.Example",
            "/my/testing/Example",
            **proxy_args
        )

    def _add_client(self, callback, *args, **kwargs):
        """Add a client."""
        self.clients.append(callback)

    def _publish_service(self):
        """Publish the service on DBus."""
        self.message_bus.publish_object(
            "/my/testing/Example",
            self.service
        )
        self.message_bus.register_service(
            "my.testing.Example"
        )

    def _run_test(self):
        """Run a test."""
        self._publish_service()

        for client in self.clients:
            client.start()

        run_loop()

        for client in self.clients:
            client.join()

    def tearDown(self):
        """Tear down the test."""
        if self.message_bus:
            self.message_bus.disconnect()

        if self.bus:
            self.bus.down()


class DBusThreadedTestCase(AbstractDBusTestCase, metaclass=ABCMeta):
    """Test DBus support with a real DBus connection and threads."""

    def _add_client(self, callback, *args, **kwargs):
        """Add a client thread."""
        thread = Thread(
            target=callback,
            args=args,
            kwargs=kwargs,
            daemon=True,
        )
        super()._add_client(thread)


class DBusSpawnedTestCase(AbstractDBusTestCase, metaclass=ABCMeta):
    """Test DBus support with a real DBus connections and spawned processes."""

    def setUp(self):
        """Set up the test."""
        super().setUp()
        self.context = multiprocessing.get_context('spawn')

    def _add_client(self, callback, *args, **kwargs):
        """Add a client process."""
        process = self.context.Process(
            name=callback.__name__,
            target=callback,
            args=(self.bus_address, *args),
            kwargs=kwargs,
            daemon=True,
        )
        super()._add_client(process)

    def _run_test(self):
        """Run a test."""
        super()._run_test()

        # Check the exit codes of the clients.
        for client in self.clients:
            msg = "{} has finished with {}".format(
                client.name,
                client.exitcode
            )
            self.assertEqual(client.exitcode, 0, msg)



#
# The common definitions
#
from dasbus.connection import SessionMessageBus
from dasbus.error import DBusError, ErrorMapper, get_error_decorator
from dasbus.identifier import DBusServiceIdentifier
from dasbus.structure import DBusData
from dasbus.typing import Str, Int

# Define the error mapper.
ERROR_MAPPER = ErrorMapper()

# Define the message bus.
SESSION_BUS = SessionMessageBus(
    error_mapper=ERROR_MAPPER
)

# Define namespaces.
REGISTER_NAMESPACE = ("org", "example", "Register")

# Define services and objects.
REGISTER = DBusServiceIdentifier(
    namespace=REGISTER_NAMESPACE,
    message_bus=SESSION_BUS
)

# The decorator for DBus errors.
dbus_error = get_error_decorator(ERROR_MAPPER)


# Define errors.
@dbus_error("InvalidUserError", namespace=REGISTER_NAMESPACE)
class InvalidUser(DBusError):
    """The user is invalid."""
    pass


# Define structures.
class User(DBusData):
    """The user data."""

    def __init__(self):
        self._name = ""
        self._age = 0

    @property
    def name(self) -> Str:
        """Name of the user."""
        return self._name

    @name.setter
    def name(self, value: Str):
        self._name = value

    @property
    def age(self) -> Int:
        """Age of the user."""
        return self._age

    @age.setter
    def age(self, value: Int):
        self._age = value



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from textwrap import dedent
from unittest.mock import Mock

from dasbus.client.handler import ClientObjectHandler, GLibClient
from dasbus.client.proxy import ObjectProxy, disconnect_proxy, InterfaceProxy
from dasbus.constants import DBUS_FLAG_NONE
from dasbus.error import ErrorMapper, DBusError, ErrorRule
from dasbus.signal import Signal
from dasbus.specification import DBusSpecification
from dasbus.typing import get_variant, get_variant_type, VariantType

import gi
gi.require_version("Gio", "2.0")
gi.require_version("GLib", "2.0")
from gi.repository import Gio, GLib


class FakeException(Exception):
    """Fake exception from DBus calls."""
    pass


class VariantTypeFactory(object):
    """Return same objects for same type strings.

    This factory allows to easily test calls with variant types.
    """

    def __init__(self):
        self._types = {}
        self._new = VariantType.new
        self._str = VariantType.__str__
        self._repr = VariantType.__repr__

    def set_up(self):
        VariantType.new = self.get
        VariantType.__str__ = VariantType.dup_string
        VariantType.__repr__ = VariantType.dup_string

    def tear_down(self):
        VariantType.new = self._new
        VariantType.__str__ = self._str
        VariantType.__repr__ = self._repr

    def get(self, type_string):
        return self._types.setdefault(type_string, self._new(type_string))


class DBusClientTestCase(unittest.TestCase):
    """Test DBus clinet support."""

    NO_REPLY = get_variant("()", ())

    def setUp(self):
        self.maxDiff = None
        self.message_bus = Mock()
        self.connection = self.message_bus.connection
        self.error_mapper = ErrorMapper()
        self.service_name = "my.service"
        self.object_path = "/my/object"
        self.handler = None
        self.proxy = None

        self.variant_type_factory = VariantTypeFactory()
        self.variant_type_factory.set_up()

    def tearDown(self):
        self.variant_type_factory.tear_down()

    def test_variant_type_factory(self):
        """Test the variant type factory."""
        self.assertEqual(str(get_variant_type("s")), "s")
        self.assertEqual(repr(get_variant_type("i")), "i")

        self.assertEqual(get_variant_type("s"), get_variant_type("s"))
        self.assertEqual(get_variant_type("i"), get_variant_type("i"))

        self.assertNotEqual(get_variant_type("b"), get_variant_type("i"))
        self.assertNotEqual(get_variant_type("s"), get_variant_type("u"))

    def _create_proxy(self, xml):
        """Create a proxy with a mocked message bus."""
        self.proxy = ObjectProxy(
            self.message_bus,
            self.service_name,
            self.object_path,
            error_mapper=self.error_mapper
        )
        self.handler = self.proxy._handler
        self.handler._specification = DBusSpecification.from_xml(xml)

    def test_introspect(self):
        """Test the introspection."""
        self._set_reply(get_variant("(s)", (dedent("""
        <node>
            <interface name="Interface">
                <method name="Method1"/>
            </interface>
        </node>
        """), )))

        self.handler = ClientObjectHandler(
            self.message_bus,
            self.service_name,
            self.object_path
        )
        self.assertIsNotNone(self.handler.specification)
        self._check_call(
            "org.freedesktop.DBus.Introspectable",
            "Introspect",
            reply_type=get_variant_type("(s)")
        )

        self.assertIn(
            DBusSpecification.Method("Method1", "Interface", None, None),
            self.handler.specification.members
        )

    def test_method(self):
        """Test the method proxy."""
        self._create_proxy("""
        <node>
            <interface name="Interface">
                <method name="Method1"/>
                <method name="Method2">
                    <arg direction="in" name="x" type="i"/>
                </method>
                <method name="Method3">
                    <arg direction="out" name="return" type="i"/>
                </method>
                <method name="Method4">
                    <arg direction="in" name="x" type="ad"/>
                    <arg direction="in" name="y" type="o"/>
                    <arg direction="out" name="return" type="(ib)"/>
                </method>
                <method name="Method5">
                    <arg direction="out" name="return_x" type="i"/>
                    <arg direction="out" name="return_y" type="i"/>
                </method>
            </interface>
        </node>
        """)

        self.assertTrue(callable(self.proxy.Method1))
        self.assertEqual(self.proxy.Method1, self.proxy.Method1)

        self._set_reply(self.NO_REPLY)
        self.assertEqual(self.proxy.Method1(), None)
        self._check_call(
            "Interface",
            "Method1"
        )

        self._set_reply(self.NO_REPLY)
        self.assertEqual(self.proxy.Method2(1), None)
        self._check_call(
            "Interface",
            "Method2",
            parameters=get_variant("(i)", (1, ))
        )

        self._set_reply(get_variant("(i)", (0, )))
        self.assertEqual(self.proxy.Method3(), 0)
        self._check_call(
            "Interface",
            "Method3",
            reply_type=get_variant_type("(i)")
        )

        self._set_reply(get_variant("((ib))", ((1, True), )))
        self.assertEqual(self.proxy.Method4([1.2, 2.3], "/my/path"),
                         (1, True))
        self._check_call(
            "Interface",
            "Method4",
            parameters=get_variant("(ado)", ([1.2, 2.3], "/my/path")),
            reply_type=get_variant_type("((ib))")
        )

        self._set_reply(get_variant("(ii)", (1, 2)))
        self.assertEqual(self.proxy.Method5(), (1, 2))
        self._check_call(
            "Interface",
            "Method5",
            reply_type=get_variant_type("(ii)")
        )

        # Handle unregistered remote exception.
        self._set_reply(Gio.DBusError.new_for_dbus_error(
            "org.test.Unknown",
            "My message."
        ))

        with self.assertRaises(DBusError) as cm:
            self.proxy.Method1()

        self.assertTrue("My message." in str(cm.exception))

        # Handle registered remote exception.
        self.error_mapper.add_rule(ErrorRule(
            exception_type=FakeException,
            error_name="org.test.Unknown"
        ))

        self._set_reply(Gio.DBusError.new_for_dbus_error(
            "org.test.Unknown",
            "My message."
        ))

        with self.assertRaises(FakeException) as cm:
            self.proxy.Method1()

        self.assertEqual(str(cm.exception), "My message.")

        # Handle timeout exception.
        self._set_reply(GLib.Error.new_literal(
            Gio.io_error_quark(),
            "Timeout was reached",
            Gio.IOErrorEnum.TIMED_OUT
        ))

        with self.assertRaises(TimeoutError) as cm:
            self.proxy.Method1()

        self.assertEqual(
            str(cm.exception),
            "The DBus call timeout was reached."
        )

        # Handle local exception.
        self._set_reply(Exception("My message."))

        with self.assertRaises(Exception) as cm:
            self.proxy.Method1()

        self.assertEqual(str(cm.exception), "My message.")

        # Test invalid method.
        with self.assertRaises(AttributeError) as cm:
            self.proxy.MethodInvalid()

        self.assertEqual(
            "DBus object has no attribute 'MethodInvalid'.",
            str(cm.exception)
        )

        # Test invalid attribute.
        with self.assertRaises(AttributeError) as cm:
            self.proxy.Method1 = lambda: 1

        self.assertEqual(
            "Can't set DBus attribute 'Method1'.",
            str(cm.exception)
        )

    def test_invalid_method_result(self):
        """Test a method proxy with an invalid result."""
        self._create_proxy("""
        <node>
            <interface name="Interface">
                <method name="Method">
                    <arg direction="out" name="return" type="t"/>
                </method>
            </interface>
        </node>
        """)

        self._set_reply(get_variant("i", -1))
        with self.assertRaises(TypeError):
            self.proxy.Method()

    def _set_reply(self, reply_value):
        """Set the reply of the DBus call."""
        self.connection.call_sync.reset_mock()

        if isinstance(reply_value, Exception):
            self.connection.call_sync.side_effect = reply_value
        else:
            self.connection.call_sync.return_value = reply_value

    def _check_call(self, interface_name, method_name, parameters=None,
                    reply_type=None):
        """Check the DBus call."""
        self.connection.call_sync.assert_called_once_with(
            self.service_name,
            self.object_path,
            interface_name,
            method_name,
            parameters,
            reply_type,
            DBUS_FLAG_NONE,
            GLibClient.DBUS_TIMEOUT_NONE,
            None
        )

        self.connection.call_sync.reset_mock()

    def test_async_method(self):
        """Test asynchronous calls of a method proxy."""
        self._create_proxy("""
        <node>
            <interface name="Interface">
                <method name="Method1"/>
                <method name="Method2">
                    <arg direction="in" name="x" type="i"/>
                    <arg direction="in" name="y" type="i"/>
                    <arg direction="out" name="return" type="i"/>
                </method>
            </interface>
        </node>
        """)
        callback = Mock()
        callback_args = ("A", "B")
        self.proxy.Method1(callback=callback, callback_args=callback_args)
        self._check_async_call(
            "Interface",
            "Method1",
            callback,
            callback_args
        )

        self._finish_async_call(self.NO_REPLY, callback, callback_args)
        callback.assert_called_once_with(None, "A", "B")

        callback = Mock()
        callback_args = ("A", "B")
        self.proxy.Method2(
            1, 2, callback=callback, callback_args=callback_args
        )
        self._check_async_call(
            "Interface",
            "Method2",
            callback,
            callback_args,
            get_variant("(ii)", (1, 2)),
            get_variant_type("(i)")
        )

        self._finish_async_call(
            get_variant("(i)", (3, )),
            callback,
            callback_args
        )
        callback.assert_called_once_with(3, "A", "B")

        self.error_mapper.add_rule(ErrorRule(
            exception_type=FakeException,
            error_name="org.test.Unknown"
        ))

        callback = Mock()
        callback_args = ("A", "B")

        error = Gio.DBusError.new_for_dbus_error(
            "org.test.Unknown",
            "My message."
        )

        with self.assertRaises(FakeException) as cm:
            self._finish_async_call(error, callback, callback_args)

        self.assertEqual(str(cm.exception), "My message.")
        callback.assert_not_called()

    def _check_async_call(self, interface_name, method_name, callback,
                          callback_args, parameters=None, reply_type=None):
        """Check the asynchronous DBus call."""
        self.connection.call.assert_called_once_with(
            self.service_name,
            self.object_path,
            interface_name,
            method_name,
            parameters,
            reply_type,
            DBUS_FLAG_NONE,
            GLibClient.DBUS_TIMEOUT_NONE,
            callback=GLibClient._async_call_finish,
            user_data=(
                self.handler._method_callback,
                (callback, callback_args)
            )
        )

        self.connection.call.reset_mock()

    def _finish_async_call(self, result, callback, callback_args):
        """Finish the asynchronous call."""
        def _call_finish(result_object):
            if isinstance(result_object, Exception):
                raise result_object

            return result_object

        def _callback(finish, *args):
            callback(finish(), *args)

        GLibClient._async_call_finish(
            source_object=Mock(call_finish=_call_finish),
            result_object=result,
            user_data=(
                self.handler._method_callback,
                (_callback, callback_args)
            )
        )

    def test_property(self):
        """Test the property proxy."""
        self._create_proxy("""
        <node>
            <interface name="Interface">
                <property name="Property1" type="i" access="readwrite" />
                <property name="Property2" type="s" access="read" />
                <property name="Property3" type="b" access="write" />
            </interface>
        </node>
        """)

        self._set_reply(self.NO_REPLY)
        self.proxy.Property1 = 10
        self._check_set_property("Property1", get_variant("i", 10))

        self._set_reply(get_variant("(v)", (get_variant("i", 20), )))
        self.assertEqual(self.proxy.Property1, 20)
        self._check_get_property("Property1")

        with self.assertRaises(AttributeError) as cm:
            self.proxy.Property2 = "World"

        self.assertEqual(str(cm.exception), "Can't set DBus property.")

        self._set_reply(get_variant("(v)", (get_variant("s", "Hello"), )))
        self.assertEqual(self.proxy.Property2, "Hello")
        self._check_get_property("Property2")

        self._set_reply(self.NO_REPLY)
        self.proxy.Property3 = False
        self._check_set_property("Property3", get_variant("b", False))

        with self.assertRaises(AttributeError) as cm:
            self.fail(self.proxy.Property3)

        self.assertEqual(str(cm.exception), "Can't read DBus property.")

        with self.assertRaises(AttributeError) as cm:
            self.proxy.PropertyInvalid = 0

        self.assertEqual(
            "DBus object has no attribute 'PropertyInvalid'.",
            str(cm.exception)
        )

        with self.assertRaises(AttributeError) as cm:
            self.fail(self.proxy.PropertyInvalid)

        self.assertEqual(
            "DBus object has no attribute 'PropertyInvalid'.",
            str(cm.exception)
        )

    def _check_set_property(self, name, value):
        """Check the DBus call that sets a property."""
        self._check_call(
            "org.freedesktop.DBus.Properties",
            "Set",
            get_variant("(ssv)", ("Interface", name, value)),
            None
        )

    def _check_get_property(self, name):
        """Check the DBus call that gets a property."""
        self._check_call(
            "org.freedesktop.DBus.Properties",
            "Get",
            get_variant("(ss)", ("Interface", name)),
            get_variant_type("(v)")
        )

    def test_signal(self):
        """Test the signal publishing."""
        self._create_proxy("""
        <node>
            <interface name="Interface">
                <signal name="Signal1" />
                <signal name="Signal2">
                    <arg direction="out" name="x" type="i"/>
                    <arg direction="out" name="y" type="s"/>
                </signal>
            </interface>
        </node>
        """)

        self.assertIsInstance(self.proxy.Signal1, Signal)
        self.assertEqual(self.proxy.Signal1, self.proxy.Signal1)

        self._check_signal("Interface", "Signal1", self.proxy.Signal1.emit)
        self._emit_signal(self.NO_REPLY, self.proxy.Signal1.emit)
        self.assertEqual(len(self.handler._subscriptions), 2)

        self._check_signal("Interface", "Signal2", self.proxy.Signal2.emit)
        self._emit_signal(get_variant("(is)", (1, "Test")),
                          self.proxy.Signal2.emit)
        self.assertEqual(len(self.handler._subscriptions), 4)

        with self.assertRaises(AttributeError) as cm:
            self.fail(self.proxy.SignalInvalid)

        self.assertEqual(
            "DBus object has no attribute 'SignalInvalid'.",
            str(cm.exception)
        )

        with self.assertRaises(AttributeError) as cm:
            self.proxy.Signal1 = self.handler._signal_factory()

        self.assertEqual(
            "Can't set DBus attribute 'Signal1'.",
            str(cm.exception)
        )

        self.proxy.Signal1.connect(Mock())
        self.proxy.Signal2.connect(Mock())

        disconnect_proxy(self.proxy)
        self.assertEqual(self.connection.signal_unsubscribe.call_count, 2)
        self.assertEqual(self.handler._subscriptions, [])
        self.assertEqual(self.proxy.Signal1._callbacks, [])
        self.assertEqual(self.proxy.Signal2._callbacks, [])

    def _check_signal(self, interface_name, signal_name, signal_callback):
        """Check the DBus signal subscription."""
        self.connection.signal_subscribe.assert_called_once_with(
            self.service_name,
            interface_name,
            signal_name,
            self.object_path,
            None,
            DBUS_FLAG_NONE,
            callback=GLibClient._signal_callback,
            user_data=(self.handler._signal_callback, (signal_callback, ))
        )
        self.connection.signal_subscribe.reset_mock()

    def _emit_signal(self, parameters, signal_callback):
        """Emit a DBus signal."""
        GLibClient._signal_callback(
            self.connection,
            None,
            self.object_path,
            None,
            None,
            parameters=parameters,
            user_data=(self.handler._signal_callback, (signal_callback,))
        )

    def test_error(self):
        """Test the error handling."""
        error = Exception("My message.")
        self.assertEqual(
            GLibClient.is_remote_error(error),
            False
        )

        error = Gio.DBusError.new_for_dbus_error(
            "org.test.Error",
            "My message."
        )
        self.assertEqual(
            GLibClient.is_remote_error(error),
            True
        )
        self.assertEqual(
            GLibClient.get_remote_error_name(error),
            "org.test.Error",
        )
        self.assertEqual(
            GLibClient.get_remote_error_message(error),
            "My message."
        )

    def _create_interface_proxy(self, xml, interface_name):
        """Create an interface proxy with a mocked message bus."""
        self.proxy = InterfaceProxy(
            self.message_bus,
            self.service_name,
            self.object_path,
            interface_name=interface_name,
            error_mapper=self.error_mapper
        )
        self.handler = self.proxy._handler
        self.handler._specification = DBusSpecification.from_xml(xml)

    def test_interface_proxy(self):
        """Test the interface proxy."""
        xml = """
        <node>
            <interface name="Interface1">
                <method name="Method1"/>
            </interface>
            <interface name="Interface2">
                <method name="Method2"/>
            </interface>
            <interface name="Interface3">
                <method name="Method1"/>
                <method name="Method2"/>
                <method name="Method3"/>
            </interface>
        </node>
        """

        # Test the first interface.
        self._create_interface_proxy(xml, "Interface1")

        # Test a valid method.
        self._set_reply(self.NO_REPLY)
        self.assertEqual(self.proxy.Method1(), None)
        self._check_call("Interface1", "Method1")

        # Test invalid methods.
        with self.assertRaises(AttributeError):
            self.proxy.Method2()

        with self.assertRaises(AttributeError):
            self.proxy.Method3()

        # Test the second interface.
        self._create_interface_proxy(xml, "Interface2")

        # Test a valid method.
        self._set_reply(self.NO_REPLY)
        self.assertEqual(self.proxy.Method2(), None)
        self._check_call("Interface2", "Method2")

        # Test invalid methods.
        with self.assertRaises(AttributeError):
            self.proxy.Method1()

        with self.assertRaises(AttributeError):
            self.proxy.Method3()

        # Test the third interface.
        self._create_interface_proxy(xml, "Interface3")

        # Test a valid method.
        self._set_reply(self.NO_REPLY)
        self.assertEqual(self.proxy.Method1(), None)
        self._check_call("Interface3", "Method1")

        self._set_reply(self.NO_REPLY)
        self.assertEqual(self.proxy.Method2(), None)
        self._check_call("Interface3", "Method2")

        self._set_reply(self.NO_REPLY)
        self.assertEqual(self.proxy.Method3(), None)
        self._check_call("Interface3", "Method3")

        # Test an invalid method.
        with self.assertRaises(AttributeError):
            self.proxy.Method4()



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from dasbus.xml import XMLParser, XMLGenerator


class XMLParserTestCase(unittest.TestCase):

    def test_is_member(self):
        """Test if the element is a member of an interface."""
        element = XMLParser.xml_to_element('<method name="MethodName" />')
        self.assertEqual(XMLParser.is_member(element), True)

        element = XMLParser.xml_to_element('<signal name="SignalName" />')
        self.assertEqual(XMLParser.is_member(element), True)

        element = XMLParser.xml_to_element(
            '<property '
            'access="PropertyAccess" '
            'name="PropertyName" '
            'type="PropertyType" />'
        )
        self.assertEqual(XMLParser.is_member(element), True)

    def test_is_interface(self):
        """Test if the element is an interface."""
        element = XMLParser.xml_to_element(
            '<interface name="InterfaceName" />'
        )
        self.assertEqual(XMLParser.is_interface(element), True)

    def test_is_signal(self):
        """Test if the element is a signal."""
        element = XMLParser.xml_to_element('<signal name="SignalName" />')
        self.assertEqual(XMLParser.is_signal(element), True)

    def test_is_method(self):
        """Test if the element is a method."""
        element = XMLParser.xml_to_element('<method name="MethodName" />')
        self.assertEqual(XMLParser.is_method(element), True)

    def test_is_property(self):
        """Test if the element is a property."""
        element = XMLParser.xml_to_element(
            '<property '
            'access="PropertyAccess" '
            'name="PropertyName" '
            'type="PropertyType" />'
        )
        self.assertEqual(XMLParser.is_property(element), True)

    def test_is_parameter(self):
        """Test if the element is a parameter."""
        element = XMLParser.xml_to_element(
            '<arg '
            'direction="ParameterDirection" '
            'name="ParameterName" '
            'type="ParameterType" />'
        )
        self.assertEqual(XMLParser.is_parameter(element), True)

    def test_has_name(self):
        """Test if the element has the specified name."""
        element = XMLParser.xml_to_element('<method name="MethodName" />')
        self.assertEqual(XMLParser.has_name(element, "MethodName"), True)
        self.assertEqual(XMLParser.has_name(element, "AnotherName"), False)

    def test_get_name(self):
        """Get the name attribute."""
        element = XMLParser.xml_to_element('<method name="MethodName" />')
        self.assertEqual(XMLParser.get_name(element), "MethodName")

    def test_get_type(self):
        """Get the type attribute."""
        element = XMLParser.xml_to_element(
            '<arg '
            'direction="ParameterDirection" '
            'name="ParameterName" '
            'type="ParameterType" />'
        )
        self.assertEqual(
            XMLParser.get_type(element),
            "ParameterType"
        )

    def test_get_access(self):
        """Get the access attribute."""
        element = XMLParser.xml_to_element(
            '<property '
            'access="PropertyAccess" '
            'name="PropertyName" '
            'type="PropertyType" />'
        )
        self.assertEqual(
            XMLParser.get_access(element),
            "PropertyAccess"
        )

    def test_get_direction(self):
        """Get the direction attribute."""
        element = XMLParser.xml_to_element(
            '<arg '
            'direction="ParameterDirection" '
            'name="ParameterName" '
            'type="ParameterType" />'
        )
        self.assertEqual(
            XMLParser.get_direction(element),
            "ParameterDirection"
        )

    def test_get_interfaces_from_node(self):
        """Get interfaces from the node."""
        element = XMLParser.xml_to_element('''
        <node>
            <interface name="A" />
            <interface name="B" />
            <interface name="C" />
        </node>
        ''')
        interfaces = XMLParser.get_interfaces_from_node(element)
        self.assertEqual(interfaces.keys(), {"A", "B", "C"})


class XMLGeneratorTestCase(unittest.TestCase):

    def _compare(self, element, xml):
        self.assertEqual(
            XMLGenerator.prettify_xml(
                XMLGenerator.element_to_xml(element)
            ),
            XMLGenerator.prettify_xml(xml)
        )

    def test_node(self):
        """Test the node element."""
        self._compare(XMLGenerator.create_node(), '<node />')

    def test_interface(self):
        """Test the interface element."""
        self._compare(
            XMLGenerator.create_interface("InterfaceName"),
            '<interface name="InterfaceName" />'
        )

    def test_parameter(self):
        """Test the parameter element."""
        self._compare(
            XMLGenerator.create_parameter(
                "ParameterName",
                "ParameterType",
                "ParameterDirection"
            ),
            '<arg '
            'direction="ParameterDirection" '
            'name="ParameterName" '
            'type="ParameterType" />'
        )

    def test_property(self):
        """Test the property element."""
        self._compare(
            XMLGenerator.create_property(
                "PropertyName",
                "PropertyType",
                "PropertyAccess"
            ),
            '<property '
            'access="PropertyAccess" '
            'name="PropertyName" '
            'type="PropertyType" />'
        )

    def test_method(self):
        """Test the method element."""
        element = XMLGenerator.create_method("MethodName")
        xml = '<method name="MethodName" />'
        self._compare(element, xml)

    def test_signal(self):
        """Test the signal element."""
        element = XMLGenerator.create_signal("SignalName")
        xml = '<signal name="SignalName" />'
        self._compare(element, xml)



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from unittest.mock import Mock

from dasbus.server.container import DBusContainer, DBusContainerError
from dasbus.server.interface import dbus_interface
from dasbus.server.publishable import Publishable
from dasbus.server.template import BasicInterfaceTemplate
from dasbus.typing import Str, ObjPath


@dbus_interface("org.Project.Object")
class MyInterface(BasicInterfaceTemplate):

    def HelloWorld(self) -> Str:
        return self.implementation.hello_world()


class MyObject(Publishable):

    def for_publication(self):
        return MyInterface(self)

    def hello_world(self):
        return "Hello World!"


class MyUnpublishable(object):
    pass


class DBusContainerTestCase(unittest.TestCase):
    """Test DBus containers."""

    def setUp(self):
        self.message_bus = Mock()
        self.container = DBusContainer(
            namespace=("org", "Project"),
            basename="Object",
            message_bus=self.message_bus
        )

    def test_set_namespace(self):
        """Test set_namespace."""
        self.container.set_namespace(("org", "Another", "Project"))

        path = self.container.to_object_path(MyObject())
        self.assertEqual(path, "/org/Another/Project/Object/1")

        path = self.container.to_object_path(MyObject())
        self.assertEqual(path, "/org/Another/Project/Object/2")

    def test_to_object_path_failed(self):
        """Test failed to_object_path."""
        with self.assertRaises(TypeError) as cm:
            self.container.to_object_path(MyUnpublishable())

        self.assertEqual(
            "Type 'MyUnpublishable' is not publishable.",
            str(cm.exception)
        )

        with self.assertRaises(DBusContainerError) as cm:
            self.container._find_object_path(MyObject())

        self.assertEqual(
            "No object path found.",
            str(cm.exception)
        )

    def test_to_object_path(self):
        """Test to_object_path."""
        obj = MyObject()
        path = self.container.to_object_path(obj)

        self.message_bus.publish_object.assert_called_once()
        published_path, published_obj = \
            self.message_bus.publish_object.call_args[0]

        self.assertEqual(path, "/org/Project/Object/1")
        self.assertEqual(path, published_path)
        self.assertIsInstance(published_obj, MyInterface)
        self.assertEqual(obj, published_obj.implementation)

        self.message_bus.reset_mock()

        self.assertEqual(self.container.to_object_path(obj), path)
        self.message_bus.publish_object.assert_not_called()

        self.assertEqual(self.container.to_object_path(obj), path)
        self.message_bus.publish_object.assert_not_called()

    def test_to_object_path_list(self):
        """Test to_object_path_list."""
        objects = [MyObject(), MyObject(), MyObject()]
        paths = self.container.to_object_path_list(objects)

        self.assertEqual(self.message_bus.publish_object.call_count, 3)

        self.assertEqual(paths, [
            "/org/Project/Object/1",
            "/org/Project/Object/2",
            "/org/Project/Object/3"
        ])

        self.message_bus.reset_mock()

        self.assertEqual(paths, self.container.to_object_path_list(objects))
        self.message_bus.publish_object.assert_not_called()

        self.assertEqual(paths, self.container.to_object_path_list(objects))
        self.message_bus.publish_object.assert_not_called()

    def test_from_object_path_failed(self):
        """Test failures."""
        with self.assertRaises(DBusContainerError) as cm:
            self.container.from_object_path(ObjPath("/org/Project/Object/1"))

        self.assertEqual(
            "Unknown object path '/org/Project/Object/1'.",
            str(cm.exception)
        )

    def test_from_object_path(self):
        """Test from_object_path."""
        obj = MyObject()
        path = self.container.to_object_path(obj)

        self.assertEqual(obj, self.container.from_object_path(path))
        self.assertEqual(path, self.container.to_object_path(obj))

        self.assertEqual(obj, self.container.from_object_path(path))
        self.assertEqual(path, self.container.to_object_path(obj))

    def test_from_object_path_list(self):
        """Test from_object_path_list."""
        objects = [MyObject(), MyObject(), MyObject()]
        paths = self.container.to_object_path_list(objects)

        self.assertEqual(objects, self.container.from_object_path_list(paths))
        self.assertEqual(paths, self.container.to_object_path_list(objects))

        self.assertEqual(objects, self.container.from_object_path_list(paths))
        self.assertEqual(paths, self.container.to_object_path_list(objects))

    def test_multiple_objects(self):
        """Test multiple objects."""
        obj = MyObject()
        path = self.container.to_object_path(obj)
        self.assertEqual(path, "/org/Project/Object/1")
        self.assertEqual(obj, self.container.from_object_path(path))
        self.message_bus.publish_object.assert_called_once()
        self.message_bus.reset_mock()

        obj = MyObject()
        path = self.container.to_object_path(obj)
        self.assertEqual(path, "/org/Project/Object/2")
        self.assertEqual(obj, self.container.from_object_path(path))
        self.message_bus.publish_object.assert_called_once()
        self.message_bus.reset_mock()

        obj = MyObject()
        path = self.container.to_object_path(obj)
        self.assertEqual(path, "/org/Project/Object/3")
        self.assertEqual(obj, self.container.from_object_path(path))
        self.message_bus.publish_object.assert_called_once()
        self.message_bus.reset_mock()



#
# Identification of DBus objects, interfaces and services
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from dasbus.namespace import get_dbus_path, get_dbus_name

__all__ = [
    'DBusInterfaceIdentifier',
    'DBusObjectIdentifier',
    'DBusServiceIdentifier'
]


class DBusBaseIdentifier(object):
    """A base identifier."""

    def __init__(self, namespace, basename=None):
        """Create an identifier.

        :param namespace: a sequence of strings
        :param basename: a string with the base name or None
        """
        if basename:
            namespace = (*namespace, basename)

        self._namespace = namespace
        self._name = get_dbus_name(*namespace)
        self._path = get_dbus_path(*namespace)

    @property
    def namespace(self):
        """DBus namespace of this object."""
        return self._namespace

    def __str__(self):
        """Return the string representation."""
        return self._name


class DBusInterfaceIdentifier(DBusBaseIdentifier):
    """Identifier of a DBus interface."""

    def __init__(self, namespace, basename=None, interface_version=None):
        """Describe a DBus interface.

        :param namespace: a sequence of strings
        :param basename: a string with the base name or None
        :param interface_version: a version of the interface
        """
        super().__init__(namespace, basename=basename)
        self._interface_version = interface_version

    def _version_to_string(self, version):
        """Convert version to a string.

        :param version: a number or None
        :return: a string
        """
        if version is None:
            return ""

        return str(version)

    @property
    def interface_name(self):
        """Full name of the DBus interface."""
        return self._name + self._version_to_string(self._interface_version)

    def __str__(self):
        """Return the string representation."""
        return self.interface_name


class DBusObjectIdentifier(DBusInterfaceIdentifier):
    """Identifier of a DBus object."""

    def __init__(self, namespace, basename=None, interface_version=None,
                 object_version=None):
        """Describe a DBus object.

        :param namespace: a sequence of strings
        :param basename: a string with the base name or None
        :param interface_version: a version of the DBus interface
        :param object_version: a version of the DBus object
        """
        super().__init__(namespace, basename=basename,
                         interface_version=interface_version)
        self._object_version = object_version

    @property
    def object_path(self):
        """Full path of the DBus object."""
        return self._path + self._version_to_string(self._object_version)

    def __str__(self):
        """Return the string representation."""
        return self.object_path


class DBusServiceIdentifier(DBusObjectIdentifier):
    """Identifier of a DBus service."""

    def __init__(self, message_bus, namespace, basename=None,
                 interface_version=None, object_version=None,
                 service_version=None):
        """Describe a DBus service.

        :param message_bus: a message bus
        :param namespace: a sequence of strings
        :param basename: a string with the base name or None
        :param interface_version: a version of the DBus interface
        :param object_version: a version of the DBus object
        :param service_version: a version of the DBus service
        """
        super().__init__(namespace, basename=basename,
                         interface_version=interface_version,
                         object_version=object_version)

        self._service_version = service_version
        self._message_bus = message_bus

    @property
    def message_bus(self):
        """Message bus of the DBus service.

        :return: a message bus
        :rtype: an instance of the MessageBus class
        """
        return self._message_bus

    @property
    def service_name(self):
        """Full name of a DBus service."""
        return self._name + self._version_to_string(self._service_version)

    def __str__(self):
        """Return the string representation."""
        return self.service_name

    def _choose_object_path(self, object_id):
        """Choose an object path."""
        if object_id is None:
            return self.object_path

        if isinstance(object_id, DBusObjectIdentifier):
            return object_id.object_path

        return object_id

    def _choose_interface_name(self, interface_id):
        """Choose an interface name."""
        if interface_id is None:
            return None

        if isinstance(interface_id, DBusInterfaceIdentifier):
            return interface_id.interface_name

        return interface_id

    def get_proxy(self, object_path=None, interface_name=None,
                  **bus_arguments):
        """Returns a proxy of the DBus object.

        If no object path is specified, we will use the object path
        of this DBus service.

        If no interface name is specified, we will use none and create
        a proxy from all interfaces of the DBus object.

        :param object_path: an object identifier or a DBus path or None
        :param interface_name: an interface identifier or a DBus name or None
        :param bus_arguments: additional arguments for the message bus
        :return: a proxy object
        """
        object_path = self._choose_object_path(object_path)
        interface_name = self._choose_interface_name(interface_name)

        return self._message_bus.get_proxy(
            self.service_name,
            object_path,
            interface_name,
            **bus_arguments
        )



#
# Run the service org.example.Register.
#
from dasbus.loop import EventLoop
from dasbus.server.interface import dbus_interface
from dasbus.server.property import emits_properties_changed
from dasbus.server.template import InterfaceTemplate
from dasbus.signal import Signal
from dasbus.typing import Structure, List
from dasbus.xml import XMLGenerator
from common import SESSION_BUS, REGISTER, User, InvalidUser


@dbus_interface(REGISTER.interface_name)
class RegisterInterface(InterfaceTemplate):
    """The DBus interface of the user register."""

    def connect_signals(self):
        """Connect the signals."""
        self.watch_property("Users", self.implementation.users_changed)

    @property
    def Users(self) -> List[Structure]:
        """The list of users."""
        return User.to_structure_list(self.implementation.users)

    @emits_properties_changed
    def RegisterUser(self, user: Structure):
        """Register a new user."""
        self.implementation.register_user(User.from_structure(user))


class Register(object):
    """The implementation of the user register."""

    def __init__(self):
        self._users = []
        self._users_changed = Signal()

    @property
    def users(self):
        """The list of users."""
        return self._users

    @property
    def users_changed(self):
        """Signal the user list change."""
        return self._users_changed

    def register_user(self, user: User):
        """Register a new user."""
        if any(u for u in self.users if u.name == user.name):
            raise InvalidUser("User {} exists.".format(user.name))

        self._users.append(user)
        self._users_changed.emit()


if __name__ == "__main__":
    # Print the generated XML specification.
    print(XMLGenerator.prettify_xml(RegisterInterface.__dbus_xml__))

    try:
        # Create the register.
        register = Register()

        # Publish the register at /org/example/Register.
        SESSION_BUS.publish_object(
            REGISTER.object_path,
            RegisterInterface(register)
        )

        # Register the service name org.example.Register.
        SESSION_BUS.register_service(
            REGISTER.service_name
        )

        # Start the event loop.
        loop = EventLoop()
        loop.run()
    finally:
        # Unregister the DBus service and objects.
        SESSION_BUS.disconnect()



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from dasbus.namespace import get_dbus_name, get_dbus_path, \
    get_namespace_from_name


class DBusNamespaceTestCase(unittest.TestCase):

    def test_dbus_name(self):
        """Test DBus path."""
        self.assertEqual(get_dbus_name(), "")
        self.assertEqual(get_dbus_name("a"), "a")
        self.assertEqual(get_dbus_name("a", "b"), "a.b")
        self.assertEqual(get_dbus_name("a", "b", "c"), "a.b.c")
        self.assertEqual(get_dbus_name("org", "freedesktop", "DBus"),
                         "org.freedesktop.DBus")

    def test_dbus_path(self):
        """Test DBus path."""
        self.assertEqual(get_dbus_path(), "/")
        self.assertEqual(get_dbus_path("a"), "/a")
        self.assertEqual(get_dbus_path("a", "b"), "/a/b")
        self.assertEqual(get_dbus_path("a", "b", "c"), "/a/b/c")
        self.assertEqual(get_dbus_path("org", "freedesktop", "DBus"),
                         "/org/freedesktop/DBus")

    def test_namespace(self):
        """Test namespaces."""
        self.assertEqual(get_namespace_from_name("a"), ("a",))
        self.assertEqual(get_namespace_from_name("a.b"), ("a", "b"))
        self.assertEqual(get_namespace_from_name("a.b.c"), ("a", "b", "c"))
        self.assertEqual(get_namespace_from_name("org.freedesktop.DBus"),
                         ("org", "freedesktop", "DBus"))



#
# DBus constants
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#

# Status codes.
DBUS_START_REPLY_SUCCESS = 1

# No flags are set.
DBUS_FLAG_NONE = 0

# System environment variable holding the DBus session address.
DBUS_STARTER_ADDRESS = "DBUS_STARTER_ADDRESS"

# Return values of org.freedesktop.DBus.RequestName.
DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER = 1
DBUS_REQUEST_NAME_REPLY_IN_QUEUE = 2
DBUS_REQUEST_NAME_REPLY_EXISTS = 3
DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER = 4

# Flags of org.freedesktop.DBus.RequestName.
DBUS_NAME_FLAG_ALLOW_REPLACEMENT = 0x1
DBUS_NAME_FLAG_REPLACE_EXISTING = 0x2
DBUS_NAME_FLAG_DO_NOT_QUEUE = 0x3



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest

from dasbus.error import ErrorMapper, DBusError, get_error_decorator, ErrorRule


class ExceptionA(Exception):
    """My testing exception A."""
    pass


class ExceptionA1(ExceptionA):
    """My testing exception A1."""
    pass


class ExceptionA2(ExceptionA):
    """My testing exception A2."""
    pass


class ExceptionB(Exception):
    """My testing exception B."""
    pass


class ExceptionC(Exception):
    """My testing exception C."""
    pass


class CustomRule(ErrorRule):
    """My custom rule for subclasses."""

    def match_type(self, exception_type):
        return issubclass(exception_type, self._exception_type)


class DBusErrorTestCase(unittest.TestCase):
    """Test the DBus error register and handler."""

    def setUp(self):
        self.error_mapper = ErrorMapper()

    def _check_type(self, error_name, expected_type):
        exception_type = self.error_mapper.get_exception_type(error_name)
        self.assertEqual(exception_type, expected_type)

    def _check_name(self, exception_type, expected_name):
        error_name = self.error_mapper.get_error_name(exception_type)
        self.assertEqual(error_name, expected_name)

    def test_decorators(self):
        """Test the error decorators."""
        dbus_error = get_error_decorator(self.error_mapper)

        @dbus_error("org.test.ErrorA")
        class DecoratedA(Exception):
            pass

        @dbus_error("ErrorB", namespace=("org", "test"))
        class DecoratedB(Exception):
            pass

        self._check_name(DecoratedA, "org.test.ErrorA")
        self._check_type("org.test.ErrorA", DecoratedA)

        self._check_name(DecoratedB, "org.test.ErrorB")
        self._check_type("org.test.ErrorB", DecoratedB)

    def test_simple_rule(self):
        """Test a simple rule."""
        self.error_mapper.add_rule(ErrorRule(
            exception_type=ExceptionA,
            error_name="org.test.ErrorA"
        ))

        self._check_name(ExceptionA, "org.test.ErrorA")
        self._check_name(ExceptionA1, "not.known.Error.ExceptionA1")
        self._check_name(ExceptionA2, "not.known.Error.ExceptionA2")

        self._check_type("org.test.ErrorA", ExceptionA)
        self._check_type("org.test.ErrorA1", DBusError)
        self._check_type("org.test.ErrorA2", DBusError)

        self._check_name(ExceptionB, "not.known.Error.ExceptionB")
        self._check_type("org.test.ErrorB", DBusError)

    def test_custom_rule(self):
        """Test a custom rule."""
        self.error_mapper.add_rule(CustomRule(
            exception_type=ExceptionA,
            error_name="org.test.ErrorA"
        ))

        self._check_name(ExceptionA, "org.test.ErrorA")
        self._check_name(ExceptionA1, "org.test.ErrorA")
        self._check_name(ExceptionA2, "org.test.ErrorA")

        self._check_type("org.test.ErrorA", ExceptionA)
        self._check_type("org.test.ErrorA1", DBusError)
        self._check_type("org.test.ErrorA2", DBusError)

        self._check_name(ExceptionB, "not.known.Error.ExceptionB")
        self._check_type("org.test.ErrorB", DBusError)

    def test_several_rules(self):
        """Test several rules."""
        self.error_mapper.add_rule(ErrorRule(
            exception_type=ExceptionA,
            error_name="org.test.ErrorA"
        ))
        self.error_mapper.add_rule(ErrorRule(
            exception_type=ExceptionB,
            error_name="org.test.ErrorB"
        ))

        self._check_name(ExceptionA, "org.test.ErrorA")
        self._check_name(ExceptionB, "org.test.ErrorB")
        self._check_name(ExceptionC, "not.known.Error.ExceptionC")

        self._check_type("org.test.ErrorA", ExceptionA)
        self._check_type("org.test.ErrorB", ExceptionB)
        self._check_type("org.test.ErrorC", DBusError)

    def test_rule_priorities(self):
        """Test the priorities of the rules."""
        self.error_mapper.add_rule(ErrorRule(
            exception_type=ExceptionA,
            error_name="org.test.ErrorA1"
        ))

        self._check_name(ExceptionA, "org.test.ErrorA1")
        self._check_type("org.test.ErrorA1", ExceptionA)
        self._check_type("org.test.ErrorA2", DBusError)

        self.error_mapper.add_rule(ErrorRule(
            exception_type=ExceptionA,
            error_name="org.test.ErrorA2"
        ))

        self._check_name(ExceptionA, "org.test.ErrorA2")
        self._check_type("org.test.ErrorA1", ExceptionA)
        self._check_type("org.test.ErrorA2", ExceptionA)

    def test_default_mapping(self):
        """Test the default error mapping."""
        self._check_name(ExceptionA, "not.known.Error.ExceptionA")
        self._check_type("org.test.ErrorB", DBusError)
        self._check_type("org.test.ErrorC", DBusError)

    def test_default_class(self):
        """Test the default class."""
        self._check_type("org.test.ErrorA", DBusError)

    def test_default_namespace(self):
        """Test the default namespace."""
        self._check_name(ExceptionA, "not.known.Error.ExceptionA")

    def test_failed_mapping(self):
        """Test the failed mapping."""
        self.error_mapper._error_rules = []

        with self.assertRaises(LookupError) as cm:
            self.error_mapper.get_error_name(ExceptionA)

        self.assertEqual(
            "No name found for 'ExceptionA'.",
            str(cm.exception)
        )

        with self.assertRaises(LookupError) as cm:
            self.error_mapper.get_exception_type("org.test.ErrorA")

        self.assertEqual(
            "No type found for 'org.test.ErrorA'.",
            str(cm.exception)
        )



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest

from unittest.mock import Mock, call

from dasbus.server.interface import dbus_interface
from dasbus.server.property import PropertiesInterface, PropertiesException, \
    PropertiesChanges, emits_properties_changed
from dasbus.signal import Signal
from dasbus.specification import DBusSpecificationError
from dasbus.typing import get_variant, Int
from dasbus.server.template import InterfaceTemplate


class DBusPropertySpecificationTestCase(unittest.TestCase):
    """Test generating properties from DBus specification."""

    def test_properties_mapping(self):
        """Test properties mapping."""

        class Interface(object):
            __dbus_xml__ = '''
            <node>
                <interface name="A">
                    <property name="A1" type="i" access="readwrite" />
                    <property name="A2" type="d" access="read" />
                    <property name="A3" type="s" access="write" />
                </interface>
                <interface name="B">
                    <property name="B1" type="i" access="readwrite" />
                    <property name="B2" type="d" access="read" />
                    <property name="B3" type="s" access="write" />
                </interface>
                <interface name="C">
                    <property name="C1" type="i" access="readwrite" />
                    <property name="C2" type="d" access="read" />
                    <property name="C3" type="s" access="write" />
                </interface>
            </node>
            '''

        changes = PropertiesChanges(Interface())
        mapping = {
            member.name: member.interface_name
            for member in changes._properties_specs.values()
        }
        expected_mapping = {
            "A1": "A", "B1": "B", "C1": "C",
            "A2": "A", "B2": "B", "C2": "C",
            "A3": "A", "B3": "B", "C3": "C",
        }

        self.assertDictEqual(mapping, expected_mapping)

    def test_invalid_properties_mapping(self):
        """Test for invalid properties."""

        class Interface(object):
            __dbus_xml__ = '''
            <node>
                <interface name="A">
                    <property name="A1" type="i" access="readwrite" />
                    <property name="A2" type="d" access="read" />
                    <property name="A3" type="s" access="write" />
                </interface>
                <interface name="B">
                    <property name="A1" type="i" access="readwrite" />
                    <property name="B2" type="d" access="read" />
                    <property name="B3" type="s" access="write" />
                </interface>
            </node>
            '''

        with self.assertRaises(DBusSpecificationError) as cm:
            PropertiesChanges(Interface())

        self.assertEqual(
            "DBus property 'A1' is defined in more than one interface.",
            str(cm.exception)
        )


class DBusPropertyTestCase(unittest.TestCase):
    """Test support for DBus properties."""

    @dbus_interface("I1")
    class Test1(PropertiesInterface):

        @property
        def A(self) -> Int:
            return 1

        @property
        def B(self) -> Int:
            return 2

    def test_report_changed(self):
        """Test reporting changed properties."""
        test1 = self.Test1()

        callback = Mock()
        test1.PropertiesChanged.connect(callback)

        test1.flush_changes()
        callback.assert_not_called()

        test1.report_changed_property("A")
        test1.flush_changes()
        callback.assert_called_once_with("I1", {
            "A": get_variant(Int, 1)
        }, [])
        callback.reset_mock()

        test1.report_changed_property("B")
        test1.flush_changes()
        callback.assert_called_once_with("I1", {
            "B": get_variant(Int, 2)
        }, [])
        callback.reset_mock()

        test1.report_changed_property("B")
        test1.report_changed_property("A")
        test1.flush_changes()
        callback.assert_called_once_with("I1", {
            "A": get_variant(Int, 1),
            "B": get_variant(Int, 2)
        }, [])
        callback.reset_mock()

    @dbus_interface("I2")
    class Test2(PropertiesInterface):

        def __init__(self):
            super().__init__()
            self._a = 1

        @property
        def a(self):
            return self._a

        @a.setter
        def a(self, a):
            self._a = a
            self.report_changed_property("A")

        @property
        def A(self) -> Int:
            return self.a

        @emits_properties_changed
        def SetA(self, a: Int):
            self.a = a

        @emits_properties_changed
        def SetADirectly(self, a: Int):
            self._a = a

        def SetANoSignal(self, a: Int):
            self.a = a

        def DoNothing(self):
            pass

    def test_emit_changed(self):
        """Test emitting changed properties."""
        test2 = self.Test2()

        callback = Mock()
        test2.PropertiesChanged.connect(callback)

        self.assertEqual(test2.A, 1)

        test2.DoNothing()
        self.assertEqual(test2.A, 1)
        callback.assert_not_called()

        test2.SetA(10)
        self.assertEqual(test2.A, 10)
        callback.assert_called_once_with("I2", {
            "A": get_variant(Int, 10)
        }, [])
        callback.reset_mock()

        test2.SetA(1000)
        self.assertEqual(test2.A, 1000)
        callback.assert_called_once_with("I2", {
            "A": get_variant(Int, 1000)
        }, [])
        callback.reset_mock()

        test2.SetADirectly(20)
        self.assertEqual(test2.A, 20)
        callback.assert_not_called()
        callback.reset_mock()

        test2.flush_changes()
        callback.assert_not_called()

        test2.SetANoSignal(200)
        self.assertEqual(test2.A, 200)
        callback.assert_not_called()

        test2.flush_changes()
        callback.assert_called_once_with("I2", {
            "A": get_variant(Int, 200)
        }, [])

    @dbus_interface("I3")
    class Test3(InterfaceTemplate):

        def connect_signals(self):
            super().connect_signals()
            self.implementation.module_properties_changed.connect(
                self.flush_changes
            )
            self.watch_property("A", self.implementation.a_changed)
            self.watch_property("B", self.implementation.b_changed)

        @property
        def A(self) -> Int:
            return self.implementation.a

        @emits_properties_changed
        def SetA(self, a: Int):
            self.implementation.set_a(a)

        @property
        def B(self) -> Int:
            return self.implementation.b

        @B.setter
        @emits_properties_changed
        def B(self, b: Int):
            self.implementation.b = b

    class Test3Implementation(object):

        def __init__(self):
            self.module_properties_changed = Signal()
            self.a_changed = Signal()
            self.b_changed = Signal()

            self._a = 1
            self._b = 2

        @property
        def a(self):
            return self._a

        def set_a(self, a):
            self._a = a
            self.a_changed.emit()

        @property
        def b(self):
            return self._b

        @b.setter
        def b(self, b):
            self._b = b
            self.b_changed.emit()

        def do_external_changes(self, a, b):
            self.set_a(a)
            self.b = b

        def do_secret_changes(self, a, b):
            self._a = a
            self._b = b

    def test_template(self):
        """Test the template with support for properties."""
        test3implementation = self.Test3Implementation()
        test3 = self.Test3(test3implementation)

        callback = Mock()
        test3.PropertiesChanged.connect(callback)

        self.assertEqual(test3.A, 1)
        self.assertEqual(test3.B, 2)

        test3.SetA(10)
        self.assertEqual(test3.A, 10)
        callback.assert_called_once_with("I3", {
            "A": get_variant(Int, 10)
        }, [])
        callback.reset_mock()

        test3.B = 20
        self.assertEqual(test3.B, 20)
        callback.assert_called_once_with("I3", {
            "B": get_variant(Int, 20)
        }, [])
        callback.reset_mock()

        test3implementation.do_external_changes(100, 200)
        self.assertEqual(test3.A, 100)
        self.assertEqual(test3.B, 200)
        callback.assert_not_called()

        test3implementation.module_properties_changed.emit()
        callback.assert_called_once_with("I3", {
            "A": get_variant(Int, 100),
            "B": get_variant(Int, 200)
        }, [])
        callback.reset_mock()

        test3implementation.do_secret_changes(1000, 2000)
        self.assertEqual(test3.A, 1000)
        self.assertEqual(test3.B, 2000)
        callback.assert_not_called()

        test3implementation.module_properties_changed.emit()
        callback.assert_not_called()

    @dbus_interface("I4")
    class Test4(InterfaceTemplate):

        def connect_signals(self):
            super().connect_signals()
            self.implementation.module_properties_changed.connect(
                self.flush_changes
            )
            self.watch_property("A", self.implementation.a_changed)

        @property
        def A(self) -> Int:
            return self.implementation.a

        @emits_properties_changed
        def SetA(self, a: Int):
            self.implementation.set_a(a)

        @emits_properties_changed
        def DoChanges(self, a: Int, b: Int):
            self.implementation.do_external_changes(a, b)

    @dbus_interface("I5")
    class Test5(Test4):

        def connect_signals(self):
            super().connect_signals()
            self.watch_property("B", self.implementation.b_changed)

        @property
        def B(self) -> Int:
            return self.implementation.b

        @B.setter
        @emits_properties_changed
        def B(self, b: Int):
            self.implementation.b = b

    class Test5Implementation(Test3Implementation):
        pass

    def test_multiple_interfaces(self):
        """Test template with multiple inheritance."""
        test5implementation = self.Test5Implementation()
        test5 = self.Test5(test5implementation)

        callback = Mock()
        test5.PropertiesChanged.connect(callback)

        self.assertEqual(test5.A, 1)
        self.assertEqual(test5.B, 2)

        test5.SetA(10)
        self.assertEqual(test5.A, 10)
        callback.assert_called_once_with("I4", {
            "A": get_variant(Int, 10)
        }, [])
        callback.reset_mock()

        test5.B = 20
        self.assertEqual(test5.B, 20)
        callback.assert_called_once_with("I5", {
            "B": get_variant(Int, 20)
        }, [])
        callback.reset_mock()

        test5.DoChanges(1, 2)
        self.assertEqual(test5.A, 1)
        self.assertEqual(test5.B, 2)
        callback.assert_has_calls([
            call("I4", {"A": get_variant(Int, 1)}, []),
            call("I5", {"B": get_variant(Int, 2)}, [])
        ], any_order=True)
        callback.reset_mock()

        test5implementation.do_external_changes(100, 200)
        self.assertEqual(test5.A, 100)
        self.assertEqual(test5.B, 200)
        callback.assert_not_called()

        test5implementation.module_properties_changed.emit()
        callback.assert_has_calls([
            call("I4", {"A": get_variant(Int, 100)}, []),
            call("I5", {"B": get_variant(Int, 200)}, [])
        ], any_order=True)
        callback.reset_mock()

        test5implementation.do_secret_changes(1000, 2000)
        self.assertEqual(test5.A, 1000)
        self.assertEqual(test5.B, 2000)
        callback.assert_not_called()

        test5implementation.module_properties_changed.emit()
        callback.assert_not_called()

    class Test6(PropertiesInterface):
        pass

    def test_invalid_class(self):
        """Test the properties interface with invalid class."""
        with self.assertRaises(DBusSpecificationError) as cm:
            self.Test6()

        self.assertEqual(
            "XML specification is not defined at '__dbus_xml__'.",
            str(cm.exception)
        )

    @dbus_interface("I7")
    class Test7(PropertiesInterface):
        pass

    def test_invalid_property(self):
        """Test the properties interface with invalid property."""
        test7 = self.Test7()

        with self.assertRaises(PropertiesException) as cm:
            test7.report_changed_property("A")

        self.assertEqual(
            "DBus object has no property 'A'.",
            str(cm.exception)
        )

    @dbus_interface("I8")
    class Test8(InterfaceTemplate):
        pass

    class Test8Implementation(object):
        pass

    def test_invalid_property_template(self):
        """Test the template with invalid property."""
        test8implementation = self.Test8Implementation()
        test8 = self.Test8(test8implementation)
        signal = Mock()

        with self.assertRaises(PropertiesException) as cm:
            test8.watch_property("A", signal)

        self.assertEqual(
            "DBus object has no property 'A'.",
            str(cm.exception)
        )

        signal.connect.assert_not_called()



# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="dasbus",
    version="1.7",
    author="Vendula Poncova",
    author_email="vponcova@redhat.com",
    description="DBus library in Python 3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='dbus glib library',
    url="https://github.com/rhinstaller/dasbus",
    packages=find_packages(include=['dasbus', 'dasbus.*']),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General"
        " Public License v2 or later (LGPLv2+)",
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.6',
)



#
# The common definitions
#
from dasbus.connection import SessionMessageBus
from dasbus.identifier import DBusServiceIdentifier

# Define the message bus.
SESSION_BUS = SessionMessageBus()

# Define services and objects.
NOTIFICATIONS = DBusServiceIdentifier(
    namespace=("org", "freedesktop", "Notifications"),
    message_bus=SESSION_BUS
)



#
# Inhibit the system suspend and hibernation.
#
import os
from dasbus.connection import SystemMessageBus
from dasbus.unix import GLibClientUnix

if __name__ == "__main__":
    # Create a representation of a system bus connection.
    bus = SystemMessageBus()

    # Create a proxy of the /org/freedesktop/login1 object
    # provided by the org.freedesktop.login1 service with
    # an enabled support for Unix file descriptors.
    proxy = bus.get_proxy(
        "org.freedesktop.login1",
        "/org/freedesktop/login1",
        client=GLibClientUnix
    )

    # Inhibit sleep by this example.
    print("Inhibit sleep by my-example.")
    fd = proxy.Inhibit(
        "sleep",
        "my-example",
        "Running an example",
        "block"
    )

    # List active inhibitors.
    print("Active inhibitors:")
    for inhibitor in sorted(proxy.ListInhibitors()):
        print("\t".join(map(str, inhibitor)))

    # Release the inhibition lock.
    print("Release the inhibition lock.")
    os.close(fd)



#
# Send a message to the chat room.
#
from common import CHAT

if __name__ == "__main__":
    # Create a proxy of the object /org/example/Chat
    # provided by the service org.example.Chat
    chat_proxy = CHAT.get_proxy()

    # Get an object path of the chat room.
    object_path = chat_proxy.FindRoom("Bob's room")
    print("Bob's room:", object_path)

    # Create a proxy of the object /org/example/Chat/Rooms/1
    # provided by the service org.example.Chat
    room_proxy = CHAT.get_proxy(object_path)

    # Send a message to the chat room.
    room_proxy.SendMessage("Hi, I am Alice!")

    # Get an object path of the chat room.
    object_path = chat_proxy.FindRoom("Alice's room")
    print("Alice's room:", object_path)

    # Create a proxy of the object /org/example/Chat/Rooms/2
    # provided by the service org.example.Chat
    room_proxy = CHAT.get_proxy(object_path)

    # Send a message to the chat room.
    room_proxy.SendMessage("I am Alice!")



#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
import unittest
from unittest.mock import patch, Mock

from dasbus.constants import DBUS_FLAG_NONE
from dasbus.client.observer import DBusObserver


class DBusObserverTestCase(unittest.TestCase):
    """Test DBus observers."""

    def _setup_observer(self, observer):
        """Set up the observer."""
        observer._service_available = Mock()
        observer._service_unavailable = Mock()
        self.assertFalse(observer.is_service_available)

    def _make_service_available(self, observer):
        """Make the service available."""
        observer._service_name_appeared_callback()
        self._test_if_service_available(observer)

    def _test_if_service_available(self, observer):
        """Test if service is available."""
        self.assertTrue(observer.is_service_available)

        observer._service_available.emit.assert_called_once_with(observer)
        observer._service_available.reset_mock()

        observer._service_unavailable.emit.assert_not_called()
        observer._service_unavailable.reset_mock()

    def _make_service_unavailable(self, observer):
        """Make the service unavailable."""
        observer._service_name_vanished_callback()
        self._test_if_service_unavailable(observer)

    def _test_if_service_unavailable(self, observer):
        """Test if service is unavailable."""
        self.assertFalse(observer.is_service_available)

        observer._service_unavailable.emit.assert_called_once_with(observer)
        observer._service_unavailable.reset_mock()

        observer._service_available.emit.assert_not_called()
        observer._service_available.reset_mock()

    def test_observer(self):
        """Test the observer."""
        observer = DBusObserver(Mock(), "SERVICE")
        self._setup_observer(observer)
        self._make_service_available(observer)
        self._make_service_unavailable(observer)

    @patch("dasbus.client.observer.Gio")
    def test_connect(self, gio):
        """Test Gio support for watching names."""
        dbus = Mock()
        observer = DBusObserver(dbus, "my.service")
        self._setup_observer(observer)

        # Connect the observer.
        observer.connect_once_available()

        # Check the call.
        gio.bus_watch_name_on_connection.assert_called_once()
        args, kwargs = gio.bus_watch_name_on_connection.call_args

        self.assertEqual(len(args), 5)
        self.assertEqual(len(kwargs), 0)
        self.assertEqual(args[0], dbus.connection)
        self.assertEqual(args[1], "my.service")
        self.assertEqual(args[2], DBUS_FLAG_NONE)

        name_appeared_closure = args[3]
        self.assertTrue(callable(name_appeared_closure))

        name_vanished_closure = args[4]
        self.assertTrue(callable(name_vanished_closure))

        # Check the subscription.
        subscription_id = gio.bus_watch_name_on_connection.return_value
        self.assertEqual(len(observer._subscriptions), 1)

        # Check the observer.
        self.assertFalse(observer.is_service_available)
        observer._service_available.emit.assert_not_called()
        observer._service_unavailable.emit.assert_not_called()

        # Call the name appeared closure.
        name_appeared_closure(dbus.connection, "my.service", "name.owner")
        self._test_if_service_available(observer)

        # Call the name vanished closure.
        name_vanished_closure(dbus.connection, "my.service")
        self._test_if_service_unavailable(observer)

        # Call the name appeared closure again.
        name_appeared_closure(dbus.connection, "my.service", "name.owner")
        self._test_if_service_available(observer)

        # Disconnect the observer.
        observer.disconnect()

        gio.bus_unwatch_name.assert_called_once_with(
            subscription_id
        )

        self._test_if_service_unavailable(observer)
        self.assertEqual(observer._subscriptions, [])



#
# Representation of a signal
#
# Copyright (C) 2019  Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#

__all__ = ["Signal"]


class Signal(object):
    """Default representation of a signal."""

    __slots__ = [
        "_callbacks",
        "__weakref__"
    ]

    def __init__(self):
        """Create a new signal."""
        self._callbacks = []

    def connect(self, callback):
        """Connect to a signal.

        :param callback: a function to register
        """
        self._callbacks.append(callback)

    def __call__(self, *args, **kwargs):
        """Emit a signal with the given arguments."""
        self.emit(*args, **kwargs)

    def emit(self, *args, **kwargs):
        """Emit a signal with the given arguments."""
        # The list of callbacks can be changed, so
        # use a copy of the list for the iteration.
        for callback in self._callbacks.copy():
            callback(*args, **kwargs)

    def disconnect(self, callback=None):
        """Disconnect from a signal.

        If no callback is specified, then all functions will
        be unregistered from the signal.

        If the specified callback isn't registered, do nothing.

        :param callback: a function to unregister or None
        """
        if callback is None:
            self._callbacks.clear()
            return

        try:
            self._callbacks.remove(callback)
        except ValueError:
            pass
